#!/usr/bin/python
from classGraph import Graph, NodeMap, NodeCover, Clause
import pdb, re, random, numpy, subprocess, datetime, os, os.path, copy

fp_print_out = None

########################################################
################# MAIN

def searchPattern(graph, edges, inv_att, parameters):            
    query_nodes = {'T_A': True, 'T_Z': True}
    neighbors_dict = graph.neighborsSub(-1, inv_att[Graph.EG].keys())

    tic = datetime.datetime.now()
    tmp_id = [0]
    min_support = len(edges)
    max_diff = 4
    plength = 0

    max_score = len(neighbors_dict)**2
    covert_edges = dict([(edge, (max_score, 1, None)) for edge in edges])

    all_covered_pairs = set()
    selected_patt = {}
    destinations = {}
    paths_list=[]
    paths_reached = []
    covert_change_sets = []
    covered_steps = []
    patterns = []
    pairs_sets = []
    base_pairs = []
    const_pairs_l = []
    tmp_reach = []
    min_support_lp = []

    #################### FINDING PATHS COVERING THE EXAMPLES
    ##########################################################

    #################### Prepare list of starting nodes with destinations
    for edgei in range(len(edges)):
        if not destinations.has_key(edges[edgei][0]):
            paths_list.append([edges[edgei][0]])
            destinations[edges[edgei][0]] = {}
        destinations[edges[edgei][0]][edges[edgei][1]] = edgei

    #################### Look for connecting paths, increasing length
    while len(paths_list) > 0 and len(paths_list) < parameters['max_paths']:
        plength += 1
        print "EXTENDING PATHS..."
        (paths_list, tmp_reach, reached) = joinLists(paths_list, neighbors_dict, plength, destinations)
        reached_pairs = set([(tr[-1], tr[0]) for tr in tmp_reach])
        print "Path length:%d, Nb pairs:%d reached out of %d, Nb paths:%d reached, %d others" % (plength, len(reached_pairs), len(edges), len(tmp_reach), len(paths_list))
        (paths_transactions, cons_keys) = enumerateConstrainedPaths(graph, tmp_reach, inv_att)

        min_support_lp.append(int(parameters['thres_support']*len(reached_pairs | all_covered_pairs)))
        if min_support_lp[-1] < parameters['min_examples']:
            min_support_lp[-1] = parameters['min_examples']

        name_in = parameters['out_rep']+("pathsitems_%s-%s.tmp" % (parameters['id_exp'], plength))
        name_out = parameters['out_rep']+("pathsfrequents_%s-%s.tmp" % (parameters['id_exp'], plength))

        print "MINING CLAUSES..."

        itemsets, cut_time = mineClauses(paths_transactions, cons_keys, tmp_reach, edges, min_support_lp[-1], parameters['max_time'], name_in, name_out)
        print "ORDERING CLAUSES..."
        covered_pairs = orderClauses(graph, itemsets, cons_keys, query_nodes, patterns, pairs_sets, base_pairs, edges, min_support_lp[-1], max_diff)


        #################### Update coverage info
        covered_steps.append(covered_pairs - all_covered_pairs)
        all_covered_pairs |= covered_pairs 
        print "Selected %s patterns, covering %d/%d pairs (%d/%d new)" % ( [ len(pattmp) for pattmp in patterns ], len(all_covered_pairs), len(edges), len(covered_steps[-1]), len(covered_pairs))

        parameters['min_support'] = int(parameters['thres_support']*len(all_covered_pairs))
        covert_change_sets.append(list(covered_steps[-1]))
        print "Cover changed:", [len(e) for e in covert_change_sets]

        if len(covert_change_sets) > parameters['try_init'] and sum([len(cp) for cp in covert_change_sets]) == 0:
        ### Not reached anything in first 2*try_extend steps, stop
            tmp_list = []
            print "Giving up, really not reaching..."
        elif sum([len(cp) for cp in covert_change_sets]) > 0 and \
                 ((len(covert_change_sets) > parameters['try_extend'] and sum([len(covert_change_sets[-sti]) for sti in range(1,parameters['try_extend']+1)]) == 0) \
                  or (len(covered_steps) > parameters['try_extend'] and sum([len(covered_steps[-sti]) for sti in range(1,parameters['try_extend']+1)]) == 0)):
        ### No new pairs reach during last try_extend steps, stop
            tmp_list = []
            print "No more trying..."
        elif parameters['alt_longer'] > 0 and len(covert_change_sets) >= parameters['alt_longer']:
        ### else, remove starting nodes the reached all their targets alt_longer steps ago
#        if len(covered_pairs) > ratio_focus*len(edges) and len(all_covered_pairs) < ratio_covers*len(edges):
            print "Discard all reached %d steps ago..." % parameters['alt_longer']
            for edge in covert_change_sets[-parameters['alt_longer']]:                    
                if destinations.has_key(edge[0]) and destinations[edge[0]].has_key(edge[1]):
                    del destinations[edge[0]][edge[1]]
                ### previously was equivalent to:
                ## destinations[edge[0]] = {}
            tmp_list = []
            for path in paths_list+tmp_reach:
                if len(destinations[path[-1]]) > 0:
                    tmp_list.append(path)
        else:
            tmp_list = paths_list + tmp_reach
            ### previously was equivalent to:
            ## tmp_list = paths_list
        paths_list = tmp_list

    toc = datetime.datetime.now()
    #################### CONSTRUCTING CLAUSES
    ##########################################################
    print "PREPARING PATHS..."
    (perfect_ids, blue_sets, idems_ids, set_inclusion) = preparePaths(patterns, pairs_sets, base_pairs)
    print "BUILDING PATTERN..."
    parameters['min_support'] = parameters['min_examples']
    if parameters['combine_paths']:
        covert_change_sets.append( selectClauses(graph, patterns, selected_patt, covert_edges, perfect_ids, blue_sets, idems_ids, set_inclusion, parameters, tmp_id))
    else:
        covert_change_sets.append( selectPClauses(graph, patterns, selected_patt, covert_edges, perfect_ids, blue_sets, idems_ids, set_inclusion, parameters, tmp_id))

    #################### SELECTION OF FINAL PATTERNS
    ##########################################################
    covert_patts = {}
    covered_blues = set()
    covered_reds = set()
    used_paths = set()
    nb_selected = 0
    essential = []
    for pair, covert in covert_edges.items():
        covert_patts.setdefault(covert[2], [[], [], [], 0])[0].append(pair)

    tmp_sel = []
    for patt in covert_patts.keys():
        if patt != None:
            covert_patts[patt][1] = selected_patt[patt].bluePairs()
            covert_patts[patt][2] = selected_patt[patt].redPairs()
            if selected_patt[patt].combination is not None:
                used_paths |= set([s[:2] for s in selected_patt[patt].combination['combining']])
            if parameters['thres_acc']*selected_patt[patt].nbRedPairs() < len(covert_patts[patt][0]) \
                   and parameters['min_examples'] < len(covert_patts[patt][0]): 
                tmp_sel.append(patt)

    inclus_patt = set()
    for bs1, patt1 in enumerate(tmp_sel):
        for bs2, patt2 in enumerate(tmp_sel[:bs1]):
            if covert_patts[patt1][1] <= covert_patts[patt2][1]:
                inclus_patt.add(patt1)
            elif covert_patts[patt2][1] <= covert_patts[patt1][1]:
                inclus_patt.add(patt2)

    i = 1
    for patt in sorted(set(tmp_sel) - inclus_patt, key=lambda x: len(covert_patts[x][1]), reverse=True):
        if parameters['limit_pattern'] < 1 or i <= parameters['limit_pattern']:
            essential.append(selected_patt[patt])
            covered_blues |= covert_patts[patt][1]
            covered_reds |= covert_patts[patt][2]
        covert_patts[patt][3] = i
        i += 1

    tac = datetime.datetime.now()
    details = {'selected_patt': selected_patt, 'covert_patts': covert_patts, 'patterns': patterns, 'used_paths': used_paths, 'covered_blues': covered_blues, 'all_covered_pairs': all_covered_pairs, 'covered_reds': covered_reds, 'tic': tic, 'toc': toc, 'tac':tac, 'nb_selected': len(essential)}
    return  essential, details 


########################################################
################# CONSTRUCTING PATHS

def joinLists(org_list, extensions_dict, cycle_check=0, destinations=None):
    exclude_ext = set()
    ext_list = []
    reached_list = []
    counts_reached = {}
    for ol in org_list:
        if cycle_check > 0:
            exclude_ext = set(ol[:cycle_check])
        for ext in (extensions_dict.get(ol[0], set()) - exclude_ext):
            if destinations != None and ext in destinations[ol[-1]].keys():
                counts_reached.setdefault(destinations[ol[-1]][ext], 0)
                counts_reached[destinations[ol[-1]][ext]] += 1
                reached_list.append([ext]+ol)
            else:
                ext_list.append([ext]+ol)
    return (ext_list, reached_list, counts_reached)

def constraintsFromIds(constraints_ids, constraints_keys):
    constraints_tmp = Graph()
    for constraint in constraints_ids:
        const = list(constraints_keys[constraint])
        if len(const[0]) == 2:
            constraints_tmp.parseElem(Graph.EG, const[0], const[1][0], const[1][1])
        else:
            constraints_tmp.parseElem(Graph.ND, const[0], const[1][0], const[1][1])
    return constraints_tmp

def enumerateConstrainedPaths(graph, paths, inv_att, preff=""):
    labels_enum = []
    cons_keys = []
    
    paths_transactions = []
    for pi in range(len(paths)):
        paths_transactions.append([])
        path = paths[pi]
        if pi == 0:
            if type(preff) == list and len(preff) == len(graph):
                nodes_labels = list(preff)
            else:
            ### Paths are built in reverse order !
                nodes_labels = ['T_%s%d' % (preff, elemi) for elemi in range(len(path)-1,-1,-1)]
                nodes_labels[0] = 'T_Z'
                nodes_labels[-1] = 'T_A'
            elemi = -1
            for elemi in range(len(path)-1): ## all paths have same length
                labels_enum.extend([nodes_labels[elemi], (nodes_labels[elemi], nodes_labels[elemi+1]), (nodes_labels[elemi+1], nodes_labels[elemi])])
                for elemj in range(elemi+1):
                    labels_enum.append((nodes_labels[elemj], nodes_labels[elemi+1]))
            labels_enum.append(nodes_labels[elemi+1])
        cp = 0
        for constraint_list in graph.getConstrainedPath(paths[pi], inv_att):
            for constraint in constraint_list:
                cons_key = (labels_enum[cp], constraint)
                if not cons_key in cons_keys:
                    cons_keys.append(cons_key)
                paths_transactions[-1].append(cons_keys.index(cons_key))
            cp += 1
    return (paths_transactions, cons_keys)

def mineClauses(paths_transactions, cons_keys, paths, edges, min_support, max_time, name_in="paths_in.tmp", name_out="paths_out.tmp"):
    connected_sets = {}
    for itemid, cons_key in enumerate(cons_keys):
        if len(cons_key[0]) == 2 and not re.match("^Na.Nb$", cons_key[1][1]):
            connected_sets.setdefault((min(cons_key[0]), max(cons_key[0])), set()).add(itemid)
    connection_list = connected_sets.values()

    fp = open(name_in, "w")
    for set_c in connection_list:
        fp.write('> '+' '.join(map(str, set_c))+'\n')
    for (tid, path_c) in enumerate(paths_transactions):
        fp.write("%s %d\n" % (' '.join(map(str, path_c)), edges.index((paths[tid][-1], paths[tid][0]))))
    fp.close()

    subprocess.call(["./mine_paths.sh", name_in, name_out, str(min_support), str(max_time)])

    itemsets = []
    fp = open(name_out)
    cut_time = None
    for line in fp:
        if cut_time == None:
            cut_time = line[0] == '1'
            continue
        partsU = re.match('^(?P<itemset>[0-9 ]*)\t(?P<freq>\d+):(?P<paths_ids>[0-9 ]*)\t(?P<freq_pairs>\d+):(?P<pairs_ids>[0-9 ]*)$', line)
        if partsU != None and int(partsU.group('freq_pairs')) >= min_support:
            itemsets.append(( map(int, partsU.group('itemset').split()), \
                             [edges[int(x)] for x in partsU.group('pairs_ids').split()]))
        itemsets.sort(key= lambda x: len(x[1]))
        del itemsets[500:]

    os.unlink(name_in)
    os.unlink(name_out)
    if os.path.exists(name_in +".log"):
        os.unlink(name_in +".log")
    return itemsets, cut_time


def orderClauses(graph, itemsets, cons_keys, query_nodes, patterns, pairs_sets, base_pairs, edges, min_support, max_diff):
    constraints_sets = []
    patterns.append([])
    for itemset, nodes_pairs in itemsets:
        pattern_data = Clause()
        pattern_data.itemset = itemset
        pattern_data.query_nodes = dict(query_nodes)
        pattern_data.constraints = constraintsFromIds(itemset, cons_keys) #, '([1-9][0-9]?)', pattern_data['id']+":") 
        pattern_edges = orderEdgesToMap(graph, pattern_data.constraints)
        pattern_data.nodes_map = mapPattern(graph, pattern_edges, pattern_data.constraints)
        pattern_data.cover = NodeCover(pattern_data.query_nodes, pattern_data.nodes_map, edges)

        if pattern_data.nodes_map is not None:
            constraints_sets.append(pattern_data)

    constraints_sets.sort(key=lambda x: (x.nbBluePairs(), x.nbConstraints()) ,reverse=True)
    covered_pairs = set()
    i = 0
    while i < len(constraints_sets):
        ### STOPPING FOR THIS LENGTH
        if constraints_sets[i].nbBluePairs() < min_support and constraints_sets[i].nbRedPairs() < 5*constraints_sets[i].nbBluePairs(): ## ACTUALLY THAT SHOULD NOT HAPPEN
            break #i = len(constraints_sets)

        psi = 0
        key_diff = None; position = None
        while psi < len(pairs_sets):
            diff_cp = constraints_sets[i].bluePairs() - pairs_sets[psi]
            diff_pc = pairs_sets[psi] - constraints_sets[i].bluePairs()
            if (len(diff_cp) + len(diff_pc)) < max_diff:
                key_diff = (tuple(diff_cp), tuple(diff_pc))
                position = psi
                psi = len(pairs_sets)+1 ## stop looking
                if base_pairs[position].has_key(key_diff):
                    pattern_idi = 0
                    ### check there is no more constrained pattern for same pairs set
                    ## ACTUALLY THAT SHOULD NOT BE NECESSARY WHEN MINING CLOSED PATTERNS
                    while key_diff != None and pattern_idi < len(base_pairs[position][key_diff]):
                        pattern_id = base_pairs[position][key_diff][pattern_idi]
                        if len(constraints_sets[i].constraints.nodesSet()) == len(patterns[pattern_id[0]][pattern_id[1]].constraints.nodesSet()) and \
                               constraints_sets[i].nbConstraints() <  patterns[pattern_id[0]][pattern_id[1]].nbConstraints() and \
                               constraints_sets[i].itemset <  patterns[pattern_id[0]][pattern_id[1]].itemset:
                            pattern_idi = len(base_pairs[position][key_diff])
                            key_diff = None
                        pattern_idi+=1                    
            psi += 1
            
        ### if no sufficient overlap
        if psi == len(pairs_sets):
            key_diff = ((),())
            pairs_sets.append(constraints_sets[i].bluePairs())
            base_pairs.append({})
            position = psi

        ### add
        if key_diff != None:
            covered_pairs |= constraints_sets[i].bluePairs()
            base_pairs[position].setdefault(key_diff, []).append((len(patterns)-1, len(patterns[-1])))
            constraints_sets[i].setId( '%d.%d' % (len(patterns)-1, len(patterns[-1])))
            patterns[-1].append(constraints_sets[i])
        i+=1
    return covered_pairs

def preparePaths(patterns, pairs_sets, base_pairs):
    blue_sets = []
    idems_ids = []
    perfect_ids = {}
    set_inclusion = {}

    for pos, diffs in enumerate(base_pairs):
        for (diff_cp, diff_pc), items_ids in diffs.items():
            blue_sets.append((set(pairs_sets[pos]) - set(diff_pc)) | set(diff_cp))
            for items_id in items_ids:
                if len(patterns[items_id[0]][items_id[1]].redPairs()) == 0:
                    perfect_ids.setdefault(len(idems_ids),[]).append(items_id)
            idems_ids.append(list(items_ids))

    sorted_set_ids = sorted(range(len(blue_sets)), key=lambda x: len(blue_sets[x]), reverse=True)
    i=0
    while i < len(sorted_set_ids):
        current_id = sorted_set_ids[i]
        j = i-1
        set_inclusion[current_id] = [set(), set(), set()] ## parents, ancestors and descendants
        while j >= 0:
            if sorted_set_ids[j] not in set_inclusion[current_id][1]:
                if blue_sets[current_id] <= blue_sets[sorted_set_ids[j]]:
                    set_inclusion[current_id][0].add(sorted_set_ids[j])
                    set_inclusion[current_id][1].add(sorted_set_ids[j])
                    set_inclusion[current_id][1] |= set_inclusion[sorted_set_ids[j]][1]
                    set_inclusion[sorted_set_ids[j]][2].add(current_id)
            j-=1
        i+=1
    return perfect_ids, blue_sets, idems_ids, set_inclusion

########################################################
################# MAPPING
def mapPattern(graph, pattern_edges, constraints, oi_check=True):
    debugP = False
    mapped_nodes = {}
    fragments = []
    expand = True
    while len(pattern_edges) > 0 and expand:
        nb_instances = max([0]+[len(mapped_nodes[ii]['pl']) for ii in mapped_nodes.keys()])
        if nb_instances > 20000:
            print "TOOO MANY INSTANCES (%d)! Giving up..." % nb_instances
            return (None, None, None, False)
        pattern_edge = pattern_edges.pop()
        if debugP: print "Instances: %d (%s)" % (max([0]+[len(mapped_nodes[ii]['pl']) for ii in mapped_nodes.keys()]), pattern_edge)
        ### for the nodes adjacent to current edge, if not yet seen, find mappable graph nodes
        for endi in [0,1]:
            if not mapped_nodes.has_key(pattern_edge[endi]):
                mapped_nodes[pattern_edge[endi]] = {'fragid':len(fragments), 'pl': graph.matchingElements(pattern_edge[endi], constraints)}
                fragments.append([pattern_edge[endi]])
            expand &= (mapped_nodes[pattern_edge[endi]]['pl'] == None) or (len(mapped_nodes[pattern_edge[endi]]['pl']) > 0)
            ### pdb.set_trace()
            if debugP:
                print "A %s" % expand
        ### else, if both nodes are in the same fragment, check which subgraph match with new edge constraint
        if mapped_nodes[pattern_edge[0]]['fragid'] == mapped_nodes[pattern_edge[1]]['fragid'] and expand: 
            keep_positions = []
            for i in range(len(mapped_nodes[pattern_edge[0]]['pl'])):
                if graph.elementMatch((mapped_nodes[pattern_edge[0]]['pl'][i], mapped_nodes[pattern_edge[1]]['pl'][i]), pattern_edge, constraints):
                    keep_positions.append(i)
            ### delete non matching ones
            if len(keep_positions) == 0: #len(mapped_nodes[pattern_edge[0]]['pl']):
                expand = False
                pdb.set_trace()
                if debugP:
                    print "B %s" % expand

            elif len(keep_positions) < len(mapped_nodes[pattern_edge[0]]['pl']):
                ##print "#KEEP_POS", len(keep_positions), len(mapped_nodes[pattern_edge[0]]['pl'])
                for node_pl in fragments[mapped_nodes[pattern_edge[0]]['fragid']]:
                    tmp_pp = [mapped_nodes[node_pl]['pl'][i] for i in keep_positions] 
                    mapped_nodes[node_pl]['pl'] = tmp_pp
                
            # elif 2*len(del_positions) > len(mapped_nodes[pattern_edge[0]]['pl']):
            #     del_positions.reverse()
            #     for node_pl in fragments[mapped_nodes[pattern_edge[0]]['fragid']]:
            #         for del_position in del_positions:
            #             mapped_nodes[node_pl]['pl'].pop(del_position)

        ### else, if the nodes are in different fragments, check which pairs of fragments can be combined with a matching edge
        else:
            new_fragid = mapped_nodes[pattern_edge[0]]['fragid']
            old_fragid = mapped_nodes[pattern_edge[1]]['fragid']    

            ### find graph matching edges
            ### careful with the order so the joining edge is a real one not just comparison
            match_edges = graph.matchingElements(pattern_edge, constraints)
            # if match_edges == None:
            #     pdb.set_trace()
            expand &= len(match_edges) > 0
            ### pdb.set_trace()
            if debugP:
                print "C %s" % expand
            if not expand:
                break
            
            ### for both fragments generate a dictionary of instances indexed by the graph node the respective edge's end is mapped to
            node_instances = [{}, {}]
            for endi in [0,1]:
                if mapped_nodes[pattern_edge[endi]]['pl'] != None :
                    for i in range(len(mapped_nodes[pattern_edge[endi]]['pl'])):
                        if node_instances[endi].has_key(mapped_nodes[pattern_edge[endi]]['pl'][i]):
                            node_instances[endi][mapped_nodes[pattern_edge[endi]]['pl'][i]].append(i)
                        else:
                            node_instances[endi][mapped_nodes[pattern_edge[endi]]['pl'][i]] = [i]

            ### NEITHER node is mapped
            if len(node_instances[0])+len(node_instances[1]) == 0:
                if fp_print_out != None:
                    fp_print_out.write("NEITHER mapped...")
                ### should be two lonely nodes:
                ### len(fragments[new_fragi])*len(fragments[old_fragi]) == 1
                tmp = zip(*match_edges)
                (mapped_nodes[pattern_edge[0]]['pl'], mapped_nodes[pattern_edge[1]]['pl']) = (list(tmp[0]), list(tmp[1]))
                mapped_nodes[pattern_edge[1]]['fragid'] = new_fragid
                
            ### BOTH nodes are mapped
            elif len(node_instances[0])*len(node_instances[1]) > 0:
                if fp_print_out != None:
                    fp_print_out.write("BOTH mapped...")
                joining = [[],[]]
                ### for each matching edge, check for both ends whether there is a mapping to corresponding pattern node
                ### if there is add the pair of instances lists to join
                for i in range(len(match_edges)):
                    for instance_Lid in node_instances[0].get(match_edges[i][0], []):
                        for instance_Rid in node_instances[1].get(match_edges[i][1], []):
                            ### Object Identity: check that no two nodes in the new instance are mapped to same graph node
                            joini = True
                            if oi_check:
                                nodes_map = [(node, 0) for node in fragments[new_fragid]] + [(node, 1) for node in fragments[old_fragid]]
                                nb_nodes = len(nodes_map)
                                positions = (instance_Lid, instance_Rid)
                                joini = ( len(set( [mapped_nodes[node_pl]['pl'][positions[side]] for (node_pl, side) in nodes_map])) == nb_nodes)
                            if joini:
                                joining[0].append(instance_Lid)
                                joining[1].append(instance_Rid)
                            # else:
                            #     pdb.set_trace()
                            #     print "Breaks OID..."

                ### join the list of instances of fragments
                for endi in [0,1]:
                    for node_pl in fragments[mapped_nodes[pattern_edge[endi]]['fragid']]:
                        tmp = [mapped_nodes[node_pl]['pl'][position] for position in joining[endi]] 
                        mapped_nodes[node_pl]['fragid'] = new_fragid
                        mapped_nodes[node_pl]['pl'] = tmp
                expand &= len(tmp) > 0
                if debugP:
                    print "D %s" % expand

            ### ONE node is mapped the other is not
            else:
                if fp_print_out != None:
                    fp_print_out.write("ONE mapped...")
                joining = []
                mapped_side = 0
                if len(node_instances[0]) == 0:
                    mapped_side = 1
                if fp_print_out != None:
                    fp_print_out.write("len(node_instances): %d, len(instances): %d, len(match_edges): %d " % ( len(node_instances[mapped_side]), len(mapped_nodes[pattern_edge[mapped_side]]['pl']), len(match_edges)))
                mapped_nodes[pattern_edge[1-mapped_side]]['pl'] = []
#                    pdb.set_trace()
                for i in range(len(match_edges)):
                    for instance_Mid in node_instances[mapped_side].get(match_edges[i][mapped_side], []):
#                            if True:
                            ### Object Identity: check that no two nodes in the new instance are mapped to same graph node                        
                        if not oi_check or \
                               not (match_edges[i][1-mapped_side] in [mapped_nodes[node_pl]['pl'][instance_Mid] for node_pl in fragments[mapped_nodes[pattern_edge[mapped_side]]['fragid']] ]):
                            mapped_nodes[pattern_edge[1-mapped_side]]['pl'].append(match_edges[i][1-mapped_side])
                            joining.append(instance_Mid)
                mapped_nodes[pattern_edge[1-mapped_side]]['fragid'] = new_fragid
                for node_pl in fragments[mapped_nodes[pattern_edge[mapped_side]]['fragid']]:
                    tmp = [mapped_nodes[node_pl]['pl'][position] for position in joining] 
                    mapped_nodes[node_pl]['fragid'] = new_fragid
                    mapped_nodes[node_pl]['pl'] = tmp
                expand &= len(tmp) > 0
                ### pdb.set_trace()
                if debugP:
                    print "E %s" % expand

            fragments[new_fragid] += fragments[old_fragid]
            fragments[old_fragid] = []

    if expand:
        return NodeMap(mapped_nodes)
    else:
        print 'Something went wrong while mapping the pattern...'
        return None

def orderEdgesToMap(graph, constraints):
    nodes_weights = {}
    edges_weights = {}
    for node in constraints.nodesSet():
        nodes_weights[node] = float('Inf')
    for (node, atts) in constraints.elems[Graph.ND].items():
        for (att, val) in atts.items():
            nodes_weights[node] = min(nodes_weights[node], len(graph.getConstraintMatch(1, att, val)))
            
    for (edge, atts) in constraints.elems[Graph.EG].items():
        edges_weights[edge] = float('Inf')
        for (att, val) in atts.items():
            if not re.match("^Na[<>=]Nb$", val):
                edges_weights[edge] = min(edges_weights[edge], len(graph.getConstraintMatch(0, att, val)))

    list_nodes = sorted(nodes_weights.items(), key = lambda x: x[1])
    list_edges = sorted(edges_weights.items(), key = lambda x: x[1])
    next_edge = None
    reached = set()
    ## Choose first
    if len(list_nodes) > 0 and list_nodes[0][1] < list_edges[0][1]:
        i = 0
        while i < len(list_edges):
            if list_nodes[0][0] in list_edges[i][0]:
                next_edge = list_edges.pop(i)[0]
                i = len(list_edges)
            i +=1
    else:
        next_edge = list_edges.pop(0)[0]
    ordered_edges = []
    while next_edge != None:
        ordered_edges.append(next_edge)
        reached |= set(next_edge)
        next_edge = None
        i = 0
        while i < len(list_edges):
            if list_edges[i][0][0] in reached and list_edges[i][0][1] in reached:
                ordered_edges.append(list_edges.pop(i)[0])
            elif next_edge==None and list_edges[i][1] != float('Inf')  \
                     and (list_edges[i][0][0] in reached or list_edges[i][0][1] in reached):
                next_edge = list_edges.pop(i)[0]
            else:
                i += 1
        
    if reached == constraints.nodesSet() and len(ordered_edges) == len(constraints.edgesSet()):
        ordered_edges.reverse()
        return ordered_edges
    pdb.set_trace()
    return []

def mapCombined(graph, combined, blues):
    pattern_edges = orderEdgesToMap(graph, combined.constraints)
    combined.nodes_map = mapPattern(graph, pattern_edges, combined.constraints)
    combined.cover = NodeCover(combined.query_nodes, combined.nodes_map, blues)
    if combined.nodes_map is None:
        return combined

    if combined.nbUnordered() == 1:
        print "Unordered to single %s..." % combined.getId()
        combined.cover.emptyBlue()
    if combined.nodes_map.nbMinMapsTo() == 1:
        print "Maps to single %s..." % combined.getId()
        combined.cover.emptyBlue()
    return combined

########################################################
################# COMBINING APPLY TRANSLATION

def combineConstraints(constraintsA, constraintsB, translateB):
    changed = False
    ###Translate edge constraints
    for key in constraintsB.elems[Graph.EG].keys():
        new_key = (translateB.get(key[0], key[0]), translateB.get(key[1], key[1]))
        if constraintsA.elems[Graph.EG].has_key(new_key):
            if constraintsA.elems[Graph.EG][new_key] != constraintsB.elems[Graph.EG][key]:
                changed = True
                constraintsA.elems[Graph.EG][new_key].update(constraintsB.elems[Graph.EG][key])
        else:
            changed = True
            constraintsA.elems[Graph.EG][new_key] =  dict(constraintsB.elems[Graph.EG][key])
    ### Translate node constraints
    for key in constraintsB.elems[Graph.ND].keys():
        new_key = translateB.get(key, key)
        if constraintsA.elems[Graph.ND].has_key(new_key):
            if constraintsA.elems[Graph.ND][new_key] != constraintsB.elems[Graph.ND][key]:
                changed = True
                constraintsA.elems[Graph.ND][new_key].update(constraintsB.elems[Graph.ND][key])
        else:
            changed = True
            constraintsA.elems[Graph.ND][new_key] = dict(constraintsB.elems[Graph.ND][key])
    return changed

########################################################
################# COMBINATION GENERATE TRANSLATION

def makeSignsEdges(nodes_map, enter_pairs, pid=0, signs={}):
    nodes_i = ['T_%s' % nn for nn in ['A'] + range(1,nodes_map.nbNodes()-1) +['Z'] ]
    if len(signs) == 0:
        signs.update( dict([(p, [{}, {}, {pid: 0}]) for p in enter_pairs]))
    else:
        for p in enter_pairs:
            signs[p][2][pid] = 0
    for x in range(nodes_map.nbInstances()):
        pair = (nodes_map.nodeInstance('T_A',x), nodes_map.nodeInstance('T_Z', x)) 
        if pair in enter_pairs:
            signs[pair][2][pid] += 1
            ii = 0
            signs[pair][0].setdefault((nodes_map.nodeInstance(nodes_i[ii], x), nodes_map.nodeInstance(nodes_i[ii+1], x)), set()).add((pid, ii))
            for ii in range(1, len(nodes_i)-1):
                signs[pair][0].setdefault((nodes_map.nodeInstance(nodes_i[ii], x), nodes_map.nodeInstance(nodes_i[ii+1], x)), set()).add((pid, ii))
                signs[pair][1].setdefault(nodes_map.nodeInstance(nodes_i[ii], x), set()).add((pid, ii))
    return signs

def generateCombTrans(patterns, idems_i, blues):
    signs = {}
    for pid, patt_id in enumerate(idems_i):
        makeSignsEdges(patterns[patt_id[0]][patt_id[1]].nodes_map, blues, pid, signs)

    all_modes = {}
    sign_dict = {}
    single_org = {}
    single_dict = {} ## for demultiple
    for pair, sign in signs.items():
        key_edges = ' '.join(sorted([ '::'.join(['%d.%s' % (p,v) for (p,v) in sorted(val)]) for val in sign[0].values()]))
        key_nodes = ' '.join(sorted([ '::'.join(['%d.%s' % (p,v) for (p,v) in sorted(val)]) for val in sign[1].values()]))
        sign_dict.setdefault(key_edges + '-' + key_nodes, []).append(pair)
    
    dict_keys = sorted(sign_dict.keys(), key=lambda x: len(x))
    inclusion = {}
    nb_orphaned = 0
    dki = len(dict_keys)-1
    while dki >= -nb_orphaned:
        covered = sign_dict[dict_keys[dki]]
        set_tmp = signs[covered[0]]
        inclusion[covered[0]] = []
        #### FIRST, FIND EQUIVALENT NODES IN THE PATTERN
        eqs_nodes = []; has_multiple = False
        for nb, rolesb in set_tmp[1].items():
            tmp_cand = [nb]
            for eq_nodes in eqs_nodes:
                if nb in eq_nodes:
                    tmp_cand = None
            if tmp_cand != None:
                edges_b = [e for e in set_tmp[0].keys() if (nb in e)]
                for na, rolesa in set_tmp[1].items():
                    copies = (na!= nb) and (rolesb == rolesa)
                    copies &= (len(edges_b) == len([e for e in set_tmp[0].keys() if (na in e)]))
                    i = 0
                    while copies and i < len(edges_b):
                        if edges_b[i][0] == nb:
                            edge_a = (na, edges_b[i][1])
                        else:
                            edge_a = (edges_b[i][0], na)
                        if set_tmp[0].get(edge_a, None) != set_tmp[0][edges_b[i]]: 
                            copies = False; 
                        i+=1
                    if copies:
                        has_multiple = True
                        tmp_cand.append(na)
                eqs_nodes.append(tmp_cand)

        if has_multiple:
            single_nodes = [n[0] for n in eqs_nodes]+list(covered[0])
            single_key_edges = ' '.join(sorted([ '::'.join(['%d.%s' % (p,v) for (p,v) in sorted(val)]) for k,val in set_tmp[0].items() if k[0] in single_nodes and k[1] in single_nodes]))
            single_key_nodes = ' '.join(sorted([ '::'.join(['%d.%s' % (p,v) for (p,v) in sorted(val)]) for k,val in set_tmp[1].items() if k in single_nodes]))
            single_dict.setdefault(single_key_edges + '-' + single_key_nodes, []).append(covered[0])
        else:
            single_org[dict_keys[dki]] = covered[0]
        # multiples_current = [(n[0], len(n)) for n in eqs_nodes if len(n) > 1]
        # for key_set in single_dict.get(single_key_edges + '-' + single_key_nodes, []):
        #     multiples_other = [(n[0], len(n)) for n in all_modes[key_set][3] if len(n) > 1]
        #     included = True
        #     for node_curr, count_curr in multiples_current:
        #         found = False
        #         for node_other, count_other in multiples_other:
        #             if signs[key_set][1][node_other] == set_tmp[1][node_curr] and count_curr <= count_other:
        #                 found = True
        #         included &= found 
        #     if included:
        #         inclusion[covered[0]].append(key_set)


        elem_modes = dict([[idems_i[pid], [{}, {}]] for (pid, count) in set_tmp[2].items() if count > 1])
        translate = dict([[(idems_i[pid][0], idems_i[pid][1], 0), {}] for (pid, count) in set_tmp[2].items() if count == 1])
        counts = [x[1] for x in sorted(set_tmp[2].items())]
        for typi in [0,1]:
            for (elem, mapto) in set_tmp[typi].items():
                for (pattmp, elemp) in mapto:
                    patti = idems_i[pattmp]
                    if elem_modes.has_key(patti):
                        elem_modes[patti][typi].setdefault(elemp, []).append(elem)
                    elif typi == 1:
                        translate[(patti[0], patti[1], 0)]['T_%d' % elemp] = 'T_%s' % elem 

        for patti, mapped in elem_modes.items():
            tmp_trans = {}
            trans_lists = [list(x) for x in mapped[0][0]]
            trans_lists_tmp = []
            for ii in range(1, len(mapped[0])):
                from_to = {}
                for x in mapped[0][ii]:
                    from_to.setdefault(x[0], []).append(x[1])
                while len(trans_lists) > 0:
                    curr = trans_lists.pop()
                    for x in from_to.get(curr[-1], []):
                        if x not in curr: ### find all way that do not go twice through same node
                            trans_lists_tmp.append(list(curr) + [x])
                        # else:
                        #     pdb.set_trace()
                trans_lists = trans_lists_tmp
                trans_lists_tmp = []
            for ti, trans_list in enumerate(trans_lists):
                tmp = dict([('T_%d' % (ei+1), 'T_%s' % elem) for ei, elem in enumerate(trans_list[1:-1])])
                translate[(patti[0], patti[1], ti)] = dict([('T_%d' % (ei+1), 'T_%s' % elem) for ei, elem in enumerate(trans_list[1:-1])])
        if len(translate) > 0:
            all_modes[covered[0]] = (translate, covered, counts, eqs_nodes)
        else:
            print "MISSED TRANSLATION ", covered[0]

        if dki == 0:
            #### FINISHED THE ORIGINALS CHECK ORPHANED
            orphan_keys = list(set(single_dict.keys()) - set(single_org.keys()))
            #print "ORPHANS: ", orphan_keys
            for orphan_key in orphan_keys:
                nb_orphaned+=1
                instance = single_dict[orphan_key][0]
                single_nodes = [n[0] for n in all_modes[instance][3]]+list(instance)
                signs[(-1, instance[0], instance[1])] = \
                           [dict([ (k,val) for k,val in signs[instance][0].items() if k[0] in single_nodes and k[1] in single_nodes]), \
                            dict([ (k,val) for k,val in signs[instance][1].items() if k in single_nodes])]
                counts = {}
                for roles in signs[(-1, instance[0], instance[1])][1].values():
                    for role in roles:
                        if not counts.has_key(role[0]):
                            counts[role[0]] = {role[1]:0}
                        if not counts[role[0]].has_key(role[1]):
                            counts[role[0]][role[1]] = 0
                        counts[role[0]][role[1]] += 1
                final_counts = dict([(k, max(v.values())) for (k,v) in counts.items()])
                signs[(-1, instance[0], instance[1])].append(final_counts)
                sign_dict[orphan_key] = [(-1, instance[0], instance[1])] 
                dict_keys.append(orphan_key)

        dki -=1

    orphan_keys = list(set(single_dict.keys()) - set(single_org.keys()))
    single_map = {}
    for (k, v) in single_org.items():
        single_map[v] = [v]
    for (k, vs) in single_dict.items():
        single_map.setdefault(single_org.get(k, None),[]).extend(vs)        
    return all_modes, single_map

def pruneIdems(idems_i, patterns, occurences=[]):
    pos =sorted(range(len(idems_i)), key=lambda x: len(patterns[idems_i[x][0]][idems_i[x][1]].itemset), reverse=True)
    pi = 0
    while pi < len(pos):
        curr_itemset = set(patterns[idems_i[pos[pi]][0]][idems_i[pos[pi]][1]].itemset)
        pj = pi
        while pj > 0:
            pj -= 1
            if idems_i[pos[pi]][0] == idems_i[pos[pj]][0] \
                   and curr_itemset <= set(patterns[idems_i[pos[pj]][0]][idems_i[pos[pj]][1]].itemset):
                do_pop = True
                for occurence in occurences:
                    if occurence[pos[pj]] < occurence[pos[pi]]:
                        do_pop = False
                if do_pop:
                    pos.pop(pi)
                    pi -=1
                    pj =-1
        pi+=1
    return pos


########################################################
################# CLAUSE SELECTION

def selectClauses(graph, patterns, selected_patt, covert_edges, perfect_ids, blue_sets, idems_ids, set_inclusion, parameters, tmp_id):

    base_blue = set(covert_edges.keys())
    perfects = True
    covert_change = True
    covert_change_set = set()
    skipping=0
    processing=0
    idems_sets = perfect_ids.keys()
    while True:
        if perfects and len(idems_sets) == 0:
            perfects = False
            idems_sets = [i for i in range(len(blue_sets)) if i not in perfect_ids.keys()]
        if len(idems_sets) == 0:
            break

        if covert_change:
            idems_sets.sort(key=lambda x: (sum([covert_edges[e][0] for e in blue_sets[x]]), sum([covert_edges[e][1] for e in blue_sets[x]])))
        covert_change = False
        idems_set = idems_sets.pop()
        idems_i = idems_ids[idems_set]
        all_i = list(idems_i)
        blues = sorted(blue_sets[idems_set], key=lambda x: covert_edges[x])
        if len(blues) >= parameters['min_support'] \
               and ( covert_edges[blues[parameters['min_support']-1]][0] > 0 \
                     or ( covert_edges[blues[parameters['min_support']-1]][0] == 0 and len(blues) > -covert_edges[blues[parameters['min_support']-1]][1])) :
            processing+=1
            if parameters['include_supers'] or not perfects:
                supers = set()
                for supid in set_inclusion[idems_set][1]:
                    supers.update(idems_ids[supid])
                all_i += list(supers)

            if parameters['prune_idems']:
                all_pos = pruneIdems(all_i, patterns)
                all_i = [a for (i, a) in enumerate(all_i) if i in all_pos] 

            stmp_id = list(tmp_id) + [0]
            tmp_id[-1] += 1
            # print "BLUES: ", blues
            all_modes, single_map = generateCombTrans(patterns, all_i, blues)                
#                if len(set(single_map.keys()) - set(all_modes.keys())) > 0: pdb.set_trace()
            singles_keys = sorted(set(single_map.keys()) & set(all_modes.keys()), key= lambda x: len(all_modes[x][3])) ### by number of nodes
            # if (5,26) in all_i:
            #     pdb.set_trace()
            #     ttt = True
            # else:
            #     ttt = False
            
            if stmp_id[0] % 50 == 0:
                print "ID---", stmp_id, len(idems_sets), len(singles_keys)
            covered_batch = set()
            batch = []
            for key in singles_keys:
                (mode_trans, covered, nb_maps, eq_nodes) = all_modes[key]
                join_covered = set()
                for ptt in single_map[key]:
                    join_covered |= set(all_modes[ptt][1])
                nb_reds_loc, nb_blues_loc, patt_loc = max([covert_edges[e] for e in join_covered if e[0] != -1])
                # print key, eq_nodes, nb_reds_loc, nb_blues_loc, covered, join_covered[key]
                stmp_id[-1] += 1
                combined = Clause()
                changed = False

                combined.combination = {}
                combined.combination['combinedp'] = []
                combined.combination['includedp'] = []

                combined.constraints = Graph()
                combined.query_nodes = dict(patterns[idems_i[0][0]][idems_i[0][1]].query_nodes)
                combined.setId('.'.join(["%s" % ti for ti in stmp_id]))

                for other_id in mode_trans.keys():
                    tmp_changed = combineConstraints(combined.constraints, patterns[other_id[0]][other_id[1]].constraints, mode_trans[other_id])
                if tmp_changed:
                    combined.combination['combinedp'].append(other_id)
                else:
                    combined.combination['includedp'].append(other_id)
                changed |= tmp_changed

                combined.combination['combining'] = mode_trans.keys()
                combined.combination['eq_nodes'] = ['T_%s' % e[0] for e in eq_nodes]
                combined.combination['eq_nodes_counts'] = {}
                nb_eq_nd = 1
                nb_nd = 0
                for eqn in eq_nodes:
                    nb_eq_nd *= len(eqn)
                    nb_nd += len(eqn)
                    for eqi in eqn:
                        combined.combination['eq_nodes_counts']['T_' + eqi] = len(eqn)

                batch.append(combined)
                #and nb_nd < parameters['max_nd'] \
                if nb_eq_nd < parameters['max_eq'] \
                       and len(join_covered) >= parameters['min_support'] \
                       and ( nb_reds_loc > 0 or ( nb_reds_loc == 0 and len(join_covered) > -nb_blues_loc)):
                    mapCombined(graph, combined, base_blue)

                    if combined.nbBluePairs() >= parameters['min_support']:
                        select_patt = []
                        for blue_pair in combined.bluePairs():
                            if combined.nbRedPairs() < covert_edges[blue_pair][0] \
                                   or ( combined.nbRedPairs() == covert_edges[blue_pair][0] \
                                        and combined.nbBluePairs() > -covert_edges[blue_pair][1]):
                                select_patt.append(blue_pair)
                        if len(select_patt) >= parameters['min_examples']:
                            for blue_pair in select_patt:
                                covert_change_set.add(blue_pair)
                                covert_edges[blue_pair] = (combined.nbRedPairs(), -combined.nbBluePairs(), tuple(stmp_id))
                            covert_change = True
                            selected_patt[tuple(stmp_id)] = combined #, mode_trans, covered, nb_maps, dbasis)
                            # covered_batch |= combined['blue_pairs']

                    if combined.nbBluePairs() > 0 and  combined.nbRedPairs()==0 and not perfects:
                        print "PERFECT COMBINATION", combined.getId() #pdb.set_trace()

        else:
            skipping+=1
    print "skipped: ",skipping, "\tprocessed: ", processing 
    return covert_change_set


def selectPClauses(graph, patterns, selected_patt, covert_edges, perfect_ids, blue_sets, idems_ids, set_inclusion, parameters, tmp_id):
    base_blue = set(covert_edges.keys())
    perfects = True
    covert_change = True
    covert_change_set = set()
    skipping=0
    processing=0
    idems_sets = perfect_ids.keys()
    while True:
        if perfects and len(idems_sets) == 0:
            perfects = False
            idems_sets = [i for i in range(len(blue_sets)) if i not in perfect_ids.keys()]
        if len(idems_sets) == 0:
            break

        if covert_change:
            idems_sets.sort(key=lambda x: (sum([covert_edges[e][0] for e in blue_sets[x]]), sum([covert_edges[e][1] for e in blue_sets[x]])))
        covert_change = False
        idems_set = idems_sets.pop()
        idems_i = idems_ids[idems_set]
        all_i = list(idems_i)
        blues = sorted(blue_sets[idems_set], key=lambda x: covert_edges[x])

        if len(blues) >= parameters['min_support'] \
               and ( covert_edges[blues[parameters['min_support']-1]][0] > 0 \
                     or ( covert_edges[blues[parameters['min_support']-1]][0] == 0 and len(blues) > -covert_edges[blues[parameters['min_support']-1]][1])) :

               # and ( covert_edges[blues[0]][0] > 0 \
               #       or ( covert_edges[blues[0]][0] == 0 and len(blues) > -covert_edges[blues[0]][1])) :

            processing+=1
            stmp_id = list(tmp_id) + [0]
            tmp_id[-1] += 1
            all_i.sort(key=lambda x: patterns[x[0]][x[1]].nbRedPairs())

            if stmp_id[0] % 50 == 0:
                print "ID---", stmp_id, len(idems_sets), len(all_i)
            covered_batch = set()
            batch = []
            for key in all_i:
                combined = patterns[key[0]][key[1]]

                batch.append(combined)
                if combined.nbBlue_pairs() >= parameters['min_support']:
                    select_patt = []
                    for blue_pair in combined.bluePairs():
                        if combined.nbRedPairs() < covert_edges[blue_pair][0] \
                               or ( combined.nbRedPairs() == covert_edges[blue_pair][0] \
                                    and combined.nbBluePairs() > -covert_edges[blue_pair][1]):
                            select_patt.append(blue_pair)
                    if len(select_patt) >= parameters['min_examples']:
                        for blue_pair in select_patt:
                            covert_change_set.add(blue_pair)
                            covert_edges[blue_pair] = (combined.nbRedPairs(), -combined.nbBluePairs(), tuple(stmp_id))
                            covert_change = True
                            selected_patt[tuple(stmp_id)] = combined #, mode_trans, covered, nb_maps, dbasis)

                if combined.nbBluePairs() > 0 and  combined.nbRedPairs()==0 and not perfects:
                    print "PERFECT COMBINATION", combined.getId() #pdb.set_trace()

        else:
            skipping+=1
    print "skipped: ",skipping, "\tprocessed: ", processing 
    return covert_change_set


########################################################
################# WRITING

def writeShortResults(parameters, edges, details):
    finals_graph = []
    hier_graph = []
    gene_graph = []
    usage_graph = []
    paths_graph = []
    paths_connect = []
    finals_connect = []
    patterns_graph = {} ## for generating bmg file    

    fs = open(parameters['out_rep']+("supp_pat_%s.bmg" % parameters['id_exp']), "w")
    for bsi, (bs1, covert1) in enumerate(details['covert_patts'].items()):
        if bs1 == None:
            hier_graph.append("# _attributes T_UNCOVERED cover=:%d: fill=100/100/100" % (len(covert1[0])))
        else:
            if covert1[3] > 0 and ( parameters['limit_pattern'] < 1 or covert1[3] <= parameters['limit_pattern']):
                base_color = 0
                hier_graph.append("# _attributes T_%s cover=%d:%d:%d prop=%d:%d:%d fill=255/%d/%d" % ('.'.join(map(str, bs1)), len(covert1[0]), len(covert1[1]), len(covert1[2]), parameters['min_examples'] < len(covert1[0]), len(covert1[2]) < len(covert1[0]), covert1[3], base_color+100*(1-len(covert1[0])/float(len(covert1[0])+ len(covert1[2]))), base_color+100*(1-len(covert1[1])/float(len(details['all_covered_pairs'])))))

                pattern  = details['selected_patt'][bs1]

                ### SUPPORT OUTPUT
                fs.write('%s\t' % pattern.getId())
                for pair in sorted(pattern.bluePairs()):
                    fs.write('%s,%s ' % (pair[0], pair[1]))
                fs.write('\t')
                for pair in sorted(pattern.redPairs()):
                    fs.write('%s,%s ' % (pair[0], pair[1]))
                fs.write('\n')
    
                ### GRAPH OUTPUT
                color = "fill=%d/%d/255 " % (200*(1-len(covert1[0])/float(len(covert1[0])+ len(covert1[2]))),
                                             200*(1-len(covert1[1])/float(len(details['all_covered_pairs']))))

                finals_graph.extend(pattern.listBMG(r'^T_([0-9A-Z:.]*)$', 'T_' + pattern.getId()+":" + r'\1', color))
                finals_connect.append("T_%s T_%s:A" % ('.'.join(map(str, bs1)), pattern.getId() ))
    
    fs.close()
    fh = open(parameters['out_rep']+("view_pat_%s.bmg" % parameters['id_exp']), "w")
    fh.write("#### SUMMARY\tEdge:%s Support:%f Blue:%d/%d/%d Red:%d Patterns:%d Time:%s/%s/%s\n" % (parameters['id_exp'], parameters['thres_support'], len(details['covered_blues']), len(details['all_covered_pairs']), len(edges), len(details['covered_reds']), details['nb_selected'], (details['toc']-details['tic']), (details['tac']-details['toc']), (details['tac']-details['tic'])))


    fh.write("#### SUPPORT INCLUSION\n")
    for line in sorted(hier_graph):
        fh.write(line+'\n')
    fh.write("#### PATTERNS GRAPHS\n")
    for line in sorted(finals_graph):
        fh.write(line+'\n')
    fh.write("#### CONNECTORS F\n")
    for line in sorted(finals_connect):
        fh.write(line+'\n')
    fh.close()

    # fp = open(out_rep+("pickle_pat_tmp%s.bmg" % cdesc_val), "w")
    # pickle.dump([(patterns[-1][kex], sets_final[kex][0], sets_final[kex][1]) for kex in selected_keys], fp)
    # fp.close()  
    print "THE END for %s, support %f. Covers %d of %d reached out of %d blue pairs, and %d red pairs, with %d patterns... Spent %s mining paths, %s combining." % (parameters['id_exp'], parameters['thres_support'], len(details['covered_blues']), len(details['all_covered_pairs']), len(edges), len(details['covered_reds']), details['nb_selected'], (details['toc']-details['tic']), (details['tac']-details['toc']))


# ########################################################
# ################# PATTERNS MANIPULATION

# def matchMaps(patterns_pair, filter_nodes=None):
#     ### first match T_A and T_Z
#     indices = []
#     same_nodeset = set(patterns_pair[0]['map'].keys()) == set(patterns_pair[1]['map'].keys()) 
#     for i in [0,1]:
#         if filter_nodes != None:
#             indices.append([])
#             for x in range(len(patterns_pair[i]['map']['T_A']['pl'])):
#                 if (patterns_pair[i]['map']['T_A']['pl'][x] in filter_nodes) and (patterns_pair[i]['map']['T_Z']['pl'][x] in filter_nodes):
#                     indices[-1].append(x)
#         else:
#             indices.append(range(len(patterns_pair[i]['map']['T_A']['pl'])))
#         indices[-1].sort(key= lambda x: (patterns_pair[i]['map']['T_A']['pl'][x], patterns_pair[i]['map']['T_Z']['pl'][x]))
#     cands = []
#     while len(indices[0]) > 0 and len(indices[1]) > 0:
#         if (patterns_pair[0]['map']['T_A']['pl'][indices[0][0]], patterns_pair[0]['map']['T_Z']['pl'][indices[0][0]]) == (patterns_pair[1]['map']['T_A']['pl'][indices[1][0]], patterns_pair[1]['map']['T_Z']['pl'][indices[1][0]]):
#             ### we are not interested in identity mappings
#             if not same_nodeset \
#                    or sum([ patterns_pair[0]['map'][nodeC]['pl'][indices[0][0]] != patterns_pair[1]['map'][nodeC]['pl'][indices[1][0]] for nodeC in patterns_pair[0]['map'].keys()]) > 0:
#                 cands.append((indices[0][0], indices[1][0]))
#             tmpi = 1
#             while tmpi < len(indices[1]) and (patterns_pair[0]['map']['T_A']['pl'][indices[0][0]], patterns_pair[0]['map']['T_Z']['pl'][indices[0][0]]) == (patterns_pair[1]['map']['T_A']['pl'][indices[1][tmpi]], patterns_pair[1]['map']['T_Z']['pl'][indices[1][tmpi]]):
#                 ### we are not interested in identity mappings
#                 if not same_nodeset \
#                        or sum([ patterns_pair[0]['map'][nodeC]['pl'][indices[0][0]] != patterns_pair[1]['map'][nodeC]['pl'][indices[1][tmpi]] for nodeC in patterns_pair[0]['map'].keys()]) > 0:
#                     cands.append((indices[0][0], indices[1][tmpi]))
#                 tmpi+=1
#             indices[0].pop(0)
#         elif (patterns_pair[0]['map']['T_A']['pl'][indices[0][0]], patterns_pair[0]['map']['T_Z']['pl'][indices[0][0]]) > (patterns_pair[1]['map']['T_A']['pl'][indices[1][0]], patterns_pair[1]['map']['T_Z']['pl'][indices[1][0]]):
#             indices[1].pop(0)
#         else:
#             indices[0].pop(0)

# #        cand_patterns_pair = {}
#     node_pairs = []
#     for nodeL in patterns_pair[0]['map'].keys():
#         if nodeL not in ['T_A', 'T_Z']:
#             for nodeR in patterns_pair[1]['map'].keys():
#                 if nodeR not in ['T_A', 'T_Z']:
#                     node_pairs.append((nodeL, nodeR))
# #                        cand_patterns_pair[(nodeL, nodeR)] = []

#     cands_matched = {}
#     node_pairs.sort()
#     for candi, (ll, rr) in enumerate(cands):
#         tmp = []
#         # if (ll, rr) in [(30, 6), (69, 6), (75, 6), (18, 11), (40, 11), (0, 4), (48, 4), (79, 4), (80, 4), (81, 4), (90, 4)]:
#         #     pdb.set_trace()
#         for i,(nodeL, nodeR) in enumerate(node_pairs):
#             if patterns_pair[0]['map'][nodeL]['pl'][ll] == patterns_pair[1]['map'][nodeR]['pl'][rr]:
#                 tmp.append(i)
#         if len(tmp) > 0:
#             cands_matched.setdefault(tuple(tmp), []).append(candi)
#     return sorted([([node_pairs[mid] for mid in matching], \
#              [cands[cid] for cid in canding],\
#              set([(patterns_pair[1]['map']['T_A']['pl'][cands[cid][1]], patterns_pair[1]['map']['T_Z']['pl'][cands[cid][1]]) for cid in canding])) \
#             for (matching, canding) in cands_matched.items()], key=lambda x: (len(x[0]), len(x[2])), reverse=True) 
#     return ([], {}, [])
